// This file was adapted to Typescript from the original
// example from https://github.com/inkle/inky/blob/858f7e545f864d114ed2bc662e0948a53461695c/app/export-for-web-template/main.js
// Copyright (c) 2016 inkle Ltd.
// Under the MIT license.

import { Compiler } from 'inkjs/compiler/Compiler';
import { ErrorType } from 'inkjs/engine/Error';

import "./style.scss";
import data from './story/main.ink';

const story = new Compiler(data).Compile();

story.onError = (message: string, type: ErrorType) => {
	document.getElementById('errors')!.innerText = `BUG: ${type}: ${message}`;
};

// Global tags - those at the top of the ink file
// We support:
//  # theme: dark
//  # author: Your Name
var globalTagTheme: string | null = null;
story.globalTags?.forEach((globalTag) => {
	let splitTag = splitPropertyTag(globalTag);
	switch (splitTag?.property) {
		case "theme":
			globalTagTheme = splitTag.val;
			break;
		case "author":
			var byline = document.querySelector('.byline')!;
			byline.innerHTML = "by " + splitTag.val;
			break;
	}
});

setupTheme(globalTagTheme);

var storyContainer = document.getElementById('story')!;
var outerScrollContainer = document.querySelector<HTMLElement>('.outerContainer')!;

// page features setup
var hasSave = loadSavePoint();
setupButtons(hasSave);

// Set initial save point
var savePoint = story.state.toJson();

// Kick off the start of the story!
continueStory(true);

// Main story processing function. Each time this is called it generates
// all the next content up as far as the next set of choices.
function continueStory(this: any, firstTime: boolean) {
	var delay = 0.0;

	// Don't over-scroll past new content
	var previousBottomEdge = firstTime ? 0 : contentBottomEdgeY();

	// Generate story text - loop through available content
	while (story.canContinue) {

		// Get ink to generate the next paragraph
		var paragraphText = story.Continue()!;

		// Any special tags included with this line
		var customClasses: string[] = [];

		story.currentTags?.forEach((tag) => {
			// Detect tags of the form "X: Y". Currently used for IMAGE and CLASS but could be
			// customised to be used for other things too.
			var splitTag = splitPropertyTag(tag);

			switch (splitTag?.property.toUpperCase()) {
				// AUDIO: src
				case "AUDIO":
					if ('audio' in this) {
						this.audio.pause();
						this.audio.removeAttribute('src');
						this.audio.load();
					}

					this.audio = new Audio(splitTag.val);
					this.audio.play();
					break;


				// AUDIOLOOP: src
				case "AUDIOLOOP":
					if ('audioLoop' in this) {
						this.audioLoop.pause();
						this.audioLoop.removeAttribute('src');
						this.audioLoop.load();
					}
					this.audioLoop = new Audio(splitTag.val);
					this.audioLoop.play();
					this.audioLoop.loop = true;
					break;

				// IMAGE: src
				case "IMAGE":
					var imageElement = document.createElement('img');
					imageElement.src = splitTag.val;
					storyContainer.appendChild(imageElement);

					imageElement.onload = () => {
						console.log(`scrollingto ${previousBottomEdge}`)
						scrollDown(previousBottomEdge)
					}

					showAfter(delay, imageElement);
					delay += 200.0;
					break;

				// LINK: url
				case "LINK":
					window.location.href = splitTag.val;
					break;


				// LINKOPEN: url
				case "LINKOPEN":
					window.open(splitTag.val);
					break;

				// BACKGROUND: src
				case "BACKGROUND":
					outerScrollContainer.style.backgroundImage = 'url(' + splitTag.val + ')';
					break;

				// CLASS: className
				case "CLASS":
					customClasses.push(splitTag.val);
					break;

				// CLEAR - removes all existing content.
				case "CLEAR":
					clearStory();
					break;

				// RESTART - clears everything and restarts the story from the beginning
				case "RESTART":
					restart();
					return;
			}
		});

		// Check if paragraphText is empty
		if (paragraphText?.trim().length == 0) {
			continue; // Skip empty paragraphs
		}

		// Create paragraph element (initially hidden)
		var paragraphElement = document.createElement('p');
		paragraphElement.innerHTML = paragraphText;
		storyContainer.appendChild(paragraphElement);

		// Add any custom classes derived from ink tags
		customClasses.forEach((klass) => {
			paragraphElement.classList.add(klass);
		});

		// Fade in paragraph after a short delay
		showAfter(delay, paragraphElement);
		delay += 200.0;
	}

	// Create HTML choices from ink choices
	story.currentChoices.forEach((choice) => {

		// Create paragraph with anchor element
		var customClasses: string[] = [];
		var isClickable = true;
		choice.tags?.forEach((choiceTag) => {
			var splitTag = splitPropertyTag(choiceTag);
			switch(splitTag?.property.toUpperCase()) {

				case "UNCLICKABLE":
					isClickable = false;
					break;


				case "CLASS":
					customClasses.push(splitTag.val);
					break;
			}
		});

		var choiceParagraphElement = document.createElement('p');
		choiceParagraphElement.classList.add("choice");

		customClasses.forEach((klass) => {
			choiceParagraphElement.classList.add(klass);
		});

		if (isClickable) {
			choiceParagraphElement.innerHTML = `<a href='#'>${choice.text}</a>`
		} else {
			choiceParagraphElement.innerHTML = `<span class='unclickable'>${choice.text}</span>`
		}
		storyContainer.appendChild(choiceParagraphElement);

		// Fade choice in after a short delay
		showAfter(delay, choiceParagraphElement);
		delay += 200.0;

		// Click on choice
		if (isClickable) {
			var choiceAnchorEl = choiceParagraphElement.querySelectorAll("a")[0];
			choiceAnchorEl.addEventListener("click", function (event) {

				// Don't follow <a> link
				event.preventDefault();

				// Extend height to fit
				// We do this manually so that removing elements and creating new ones doesn't
				// cause the height (and therefore scroll) to jump backwards temporarily.
				storyContainer.style.height = contentBottomEdgeY() + "px";

				// Remove all existing choices
				removeAll(".choice");

				// Tell the story where to go next
				story.ChooseChoiceIndex(choice.index);

				// This is where the save button will save from
				savePoint = story.state.toJson();

				// Aaand loop
				continueStory(false);
			});
		}
	});

	// Unset storyContainer's height, allowing it to resize itself
	storyContainer.style.height = "";

	if (!firstTime)
		scrollDown(previousBottomEdge);
}

function clearStory() {
	removeAll("p");
	removeAll("img");

	// Comment out this line if you want to leave the header visible when clearing
	setVisible(".header", false);
}

function restart() {
	clearStory();
	story.ResetState();

	setVisible(".header", true);

	// set save point to here
	savePoint = story.state.toJson();
	continueStory(true);

	outerScrollContainer.scrollTo(0, 0);
}

// -----------------------------------
// Various Helper functions
// -----------------------------------

// Detects whether the user accepts animations
function isAnimationEnabled() {
	return window.matchMedia('(prefers-reduced-motion: no-preference)').matches;
}

// Fades in an element after a specified delay
function showAfter(delay: number, el: Element) {
	if (isAnimationEnabled()) {
		el.classList.add("hide");
		setTimeout(function () { el.classList.remove("hide") }, delay);
	} else {
		// If the user doesn't want animations, show immediately
		el.classList.remove("hide");
	}
}

// Scrolls the page down, but no further than the bottom edge of what you could
// see previously, so it doesn't go too far.
function scrollDown(previousBottomEdge: number) {
	// If the user doesn't want animations, let them scroll manually
	if (!isAnimationEnabled()) {
		return;
	}

	// Line up top of screen with the bottom of where the previous content ended
	var target = previousBottomEdge;

	// Can't go further than the very bottom of the page
	var limit = outerScrollContainer.scrollHeight - outerScrollContainer.clientHeight;
	if (target > limit) target = limit;

	var start = outerScrollContainer.scrollTop;

	var dist = target - start;
	var duration = 300 + 300 * dist / 100;
	var startTime: EpochTimeStamp | null = null;
	function step(time: EpochTimeStamp) {
		if (startTime == null) startTime = time;
		var t = (time - startTime) / duration;
		var lerp = 3 * t * t - 2 * t * t * t; // ease in/out
		outerScrollContainer.scrollTo(0, (1.0 - lerp) * start + lerp * target);
		if (t < 1) requestAnimationFrame(step);
	}
	requestAnimationFrame(step);
}

// The Y coordinate of the bottom end of all the story content, used
// for growing the container, and deciding how far to scroll.
function contentBottomEdgeY() {
	var bottomElement = storyContainer.lastElementChild as HTMLElement;
	return bottomElement ? bottomElement.offsetTop + bottomElement.offsetHeight : 0;
}

// Remove all elements that match the given selector. Used for removing choices after
// you've picked one, as well as for the CLEAR and RESTART tags.
function removeAll(selector: string) {
	storyContainer.querySelectorAll<HTMLElement>(selector).forEach((el) => {
		el.parentNode!.removeChild(el);
	});
}

// Used for hiding and showing the header when you CLEAR or RESTART the story respectively.
function setVisible(selector: string, visible: boolean) {
	var allElements = storyContainer.querySelectorAll(selector);
	for (var i = 0; i < allElements.length; i++) {
		var el = allElements[i];
		if (!visible) {
			el.classList.add("invisible");
		} else {
			el.classList.remove("invisible");
		}
	}
}

// Helper for parsing out tags of the form:
//  # PROPERTY: valuedi
// e.g. IMAGE: source path
function splitPropertyTag(tag: string) {
	var propertySplitIdx = tag.indexOf(":");
	if (propertySplitIdx != null) {
		var property = tag.slice(0, propertySplitIdx).trim();
		var val = tag.slice(propertySplitIdx + 1).trim();
		return {
			property: property,
			val: val
		};
	}

	return null;
}

// Loads save state if exists in the browser memory
function loadSavePoint() {

	try {
		let savedState = window.localStorage.getItem('save-state');
		if (savedState) {
			story.state.LoadJson(savedState);
			return true;
		}
	} catch (e) {
		console.debug("Couldn't load save state");
	}
	return false;
}

// Detects which theme (light or dark) to use
function setupTheme(globalTagTheme: string | null) {
	// load theme from browser memory
	let savedTheme;
	try {
		savedTheme = window.localStorage.getItem('theme');
	} catch (e) {
		console.debug("Couldn't load saved theme");
	}

	// Check whether the OS/browser is configured for dark mode
	var browserDark = window.matchMedia("(prefers-color-scheme: dark)").matches;

	let preferredTheme = savedTheme ?? globalTagTheme ?? (browserDark ? "dark" : "light");
	document.body.setAttribute("data-theme", preferredTheme);
	console.log(`THEMES: ${savedTheme} ${globalTagTheme} ${browserDark}`)
}

// Used to hook up the functionality for global functionality buttons
function setupButtons(hasSave: boolean) {

	let rewindEl = document.getElementById("rewind");
	if (rewindEl) rewindEl.addEventListener("click", function (event) {
		removeAll("p");
		removeAll("img");
		setVisible(".header", false);
		restart();
	});

	let saveEl = document.getElementById("save");
	if (saveEl) saveEl.addEventListener("click", function (event) {
		try {
			window.localStorage.setItem('save-state', savePoint);
			document.getElementById("reload")!.removeAttribute("disabled");
			window.localStorage.setItem('theme', document.body.classList.contains("dark") ? "dark" : "");
		} catch (e) {
			console.warn("Couldn't save state");
		}

	});

	let reloadEl = document.getElementById("reload")!;
	if (!hasSave) {
		reloadEl.setAttribute("disabled", "disabled");
	}
	reloadEl.addEventListener("click", function (event) {
		if (reloadEl.getAttribute("disabled"))
			return;

		removeAll("p");
		removeAll("img");
		try {
			let savedState = window.localStorage.getItem('save-state');
			if (savedState) story.state.LoadJson(savedState);
		} catch (e) {
			console.debug("Couldn't load save state");
		}
		continueStory(true);
	});

	let themeSwitchEl = document.getElementById("theme-switch")!;
	themeSwitchEl.addEventListener("click", (_) => {
		// toggle theme between light and dark.
		let currentTheme = document.body.getAttribute("data-theme") ?? "light";
		switch(currentTheme) {
			case "light":
			default:
				currentTheme = "dark";
				break;
			case "dark":
				currentTheme = "light";
				break;
		}
		window.localStorage.setItem("theme", currentTheme);
		setupTheme(null);
	});
}
